<?php

namespace FiloBlu\CheckoutExtend\Observer;

use Magento\Framework\Event\ObserverInterface;
use Magento\Framework\Event\Observer;
use Exception;
use FiloBlu\CheckoutExtend\Helper\Data;
use Magento\Checkout\Model\Session as CheckoutSession;
use Magento\Customer\Api\Data\CustomerInterfaceFactory;
use Magento\Framework\Event\ManagerInterface;
use Magento\Newsletter\Model\SubscriberFactory;
use Magento\Quote\Model\QuoteFactory;
use Magento\Quote\Model\QuoteRepository;
use Magento\Sales\Api\OrderRepositoryInterface;
use Magento\SalesRule\Model\Coupon;
use Magento\SalesRule\Model\ResourceModel\Coupon\Usage;
use Magento\Customer\Api\Data\AddressInterface;
use Magento\SalesRule\Model\Rule\CustomerFactory;
use Psr\Log\LoggerInterface;
use Magento\Customer\Api\AccountManagementInterface;
use Magento\Sales\Model\Order\OrderCustomerExtractor;
use Magento\Customer\Model\Session;
use Magento\Sales\Api\OrderCustomerManagementInterface;
use Magento\Sales\Api\Data\OrderAddressInterface;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Quote\Model\Quote\Address as QuoteAddress;
use Magento\Quote\Model\Quote\AddressFactory as QuoteAddressFactory;
use Magento\Framework\App\Http\Context;
use Magento\Customer\Model\CustomerFactory as CustomerModelFactory;


class OnePageControllerSuccessAction implements ObserverInterface {

    /**
     * @var SubscriberFactory
     */
    protected $_subscriberFactory;
    /**
     * @var Data
     */
    protected $_helper;
    /**
     * @var QuoteFactory
     */
    protected $_quoteFactory;
    /**
     * @var CheckoutSession
     */
    protected $_checkoutSession;
    /**
     * @var CustomerInterfaceFactory
     */
    protected $customerFactory;
    /**
     * @var OrderRepositoryInterface
     */
    protected $orderRepository;
    /**
     * @var ManagerInterface
     */
    protected $_eventManager;
    /**
     * @var Coupon
     */
    protected $coupon;
    /**
     * @var Usage
     */
    protected $couponUsage;
    /**
     * @var LoggerInterface
     */
    protected $logger;
    /**
     * @var CustomerFactory
     */
    protected $ruleCustomerFactory;

    /**
     * @var QuoteRepository
     */
    private $quoteRepository;
    /**
     * @var AccountManagementInterface
     */
    private $accountManagement;
    /**
     * @var OrderCustomerExtractor
     */
    private $customerExtractor;
    /**
     * @var Session
     */
    private $customerSession;
    /**
     * @var OrderCustomerManagementInterface
     */
    private $orderCustomerManagement;
    /**
     * @var QuoteAddressFactory
     */
    private $addressFactory;
    /**
     * @var Context
     */
    private $httpContext;
    /**
     * @var CustomerModelFactory
     */
    private $customerModelFactory;

    /**
     * OnePageControllerSuccessAction constructor.
     * @param SubscriberFactory $subscriberFactory
     * @param Data $helper
     * @param QuoteFactory $quoteFactory
     * @param CheckoutSession $checkoutSession
     * @param CustomerInterfaceFactory $customerFactory
     * @param OrderRepositoryInterface $orderRepository
     * @param ManagerInterface $eventManager
     * @param Coupon $coupon
     * @param Usage $couponUsage
     * @param CustomerFactory $ruleCustomerFactory
     * @param LoggerInterface $logger
     * @param AccountManagementInterface $accountManagement
     * @param QuoteRepository $quoteRepository
     * @param OrderCustomerExtractor $customerExtractor
     * @param Session $customerSession
     * @param OrderCustomerManagementInterface $orderCustomerManagement
     * @param QuoteAddressFactory $addressFactory
     * @param Context $httpContext
     * @param CustomerModelFactory $customerModelFactory
     */
    public function __construct(
        SubscriberFactory $subscriberFactory,
        Data $helper,
        QuoteFactory $quoteFactory,
        CheckoutSession $checkoutSession,
        CustomerInterfaceFactory $customerFactory,
        OrderRepositoryInterface $orderRepository,
        ManagerInterface $eventManager,
        Coupon $coupon,
        Usage $couponUsage,
        CustomerFactory $ruleCustomerFactory,
        LoggerInterface $logger,
        AccountManagementInterface $accountManagement,
        QuoteRepository $quoteRepository,
        OrderCustomerExtractor $customerExtractor,
        Session  $customerSession,
        OrderCustomerManagementInterface $orderCustomerManagement,
        QuoteAddressFactory $addressFactory,
        Context $httpContext,
        CustomerModelFactory $customerModelFactory
    )
    {
        $this->_subscriberFactory = $subscriberFactory;
        $this->_helper = $helper;
        $this->_quoteFactory = $quoteFactory;
        $this->_checkoutSession = $checkoutSession;
        $this->customerFactory = $customerFactory;
        $this->orderRepository = $orderRepository;
        $this->_eventManager = $eventManager;
        $this->coupon = $coupon;
        $this->couponUsage = $couponUsage;
        $this->ruleCustomerFactory = $ruleCustomerFactory;
        $this->logger = $logger;
        $this->quoteRepository = $quoteRepository;
        $this->accountManagement = $accountManagement;
        $this->customerExtractor = $customerExtractor;
        $this->customerSession = $customerSession;
        $this->orderCustomerManagement = $orderCustomerManagement;
        $this->addressFactory = $addressFactory;
        $this->httpContext = $httpContext;
        $this->customerModelFactory = $customerModelFactory;
    }

    /**
     * @param Observer $observer
     */
    public function execute(Observer $observer) {
        try{
            $order = $this->_checkoutSession->getLastRealOrder();
            //convert Guest to Customer
            if(!$order->getCustomer()) {
                $quote = $this->_quoteFactory->create()->loadByIdWithoutStore($order->getQuoteId());
                $customer = $this->customerFactory->create();
                $customer->setStoreId($order->getStoreId());
                if ($order->getId() && !$customer->getId() && $this->isCheckoutRegistrationEnabled()) {
                    //create Account
                    $customerPwd = $this->_checkoutSession->getRegistrationPwd();

                    $customer = $this->customerExtractor->extract($order->getId());
                    /** @var AddressInterface[] $filteredAddresses */
                    $filteredAddresses = [];
                    foreach ($customer->getAddresses() as $address) {
                        if ($this->needToSaveAddress($order, $address)) {
                            $filteredAddresses[] = $address;
                        }
                    }
                    $customer->setAddresses($filteredAddresses);

                    $account = $this->accountManagement->createAccount($customer, $customerPwd);
                    $order->setCustomerId($account->getId());
                    $order->setCustomerIsGuest(0);
                    $this->_eventManager->dispatch(
                        'customer_register_success_checkout',
                        ['account_controller' => null, 'customer' => $account]
                    );
                    //Writes Coupon Usages
                    if (!empty($order->getCouponCode())) {
                        $this->updateCouponUsages($order, (int)$account->getId());
                    }

                    //Writes Rule usages
                    if (!empty($order->getAppliedRuleIds())) {
                        $this->addCustomerRule($order, (int)$account->getId());
                    }

                    // Subscribe to Newsletter
                    if ($this->isCheckoutNewsletterEnabled()) {
                        $this->_subscriberFactory->create()->subscribeCustomerById($account->getId());
                    }
                }
                //perform extra action on order
                    //save invoice and DNI
                    if ($this->isDniEnabled()) {
                        if ($quote) {
                            $order->setCheckoutInvoice($quote->getCheckoutInvoice());
                            $order->setCheckoutInvoiceFiscalCode($quote->getCheckoutInvoiceFiscalCode());
                            $order->setCheckoutInvoiceVatNumber($quote->getCheckoutInvoiceVatNumber());
                            $order->setCheckoutInvoiceBusinessName($quote->getCheckoutInvoiceBusinessName());
                            $order->setCheckoutInvoicePec($quote->getCheckoutInvoicePec());
                            $order->setCheckoutInvoiceSdi($quote->getCheckoutInvoiceSdi());
                        }

                        if ($this->_helper->getConfigValue("filoblu_checkoutextend_section/dni/enabled") && $quote->getDni()) {
                            $order->setDni($quote->getDni());
                        }
                    }
                    $this->orderRepository->save($order);
                    //logIn new customer

                    $newQuote = $this->_quoteFactory->create()->setCustomer($account)->setIsActive(true)->save();
                    $this->_checkoutSession->setQuoteId($newQuote->getId());
                    $this->customerSession->setCustomerId($account->getId());
                    $this->httpContext->setValue(\Magento\Customer\Model\Context::CONTEXT_AUTH, true, false);
                    $this->customerSession->setCustomerData($account);
                    $customerModel = $this->customerModelFactory->create()->updateData($account);
                    $this->customerSession->setCustomer($customerModel);
                    $this->customerSession->regenerateId();
            }

        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
        }
    }

    /**
     * @return bool
     */
    public function isCheckoutRegistrationEnabled()
    {
        try {
            $quote = $this->quoteRepository->get($this->_checkoutSession->getLastRealOrder()->getQuoteId());
            return ($this->_helper->getConfigValue("filoblu_checkoutextend_section/registration/enabled") &&  $quote->getCheckoutRegistration());
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return false;
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
            return false;
        }

    }

    /**
     * @return bool
     */
    public function isCheckoutNewsletterEnabled()
    {
        try{
            $quote = $this->quoteRepository->get($this->_checkoutSession->getLastRealOrder()->getQuoteId());
            return ($this->_helper->getConfigValue("filoblu_checkoutextend_section/newsletter/enabled") && $quote->getIsSubscribeForNewsletter());
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return false;
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
            return false;
        }
    }

    /**
     * @return bool
     */
    public function isDniEnabled(){
        try{
            $quote = $this->quoteRepository->get($this->_checkoutSession->getLastRealOrder()->getQuoteId());
            return ($this->_helper->getConfigValue("filoblu_checkoutextend_section/invoice/enabled") && $quote->getCheckoutInvoice());
        }catch (\Exception $e){
            $this->logger->critical($e->getMessage());
            return false;
        }catch (\Throwable $t){
            $this->logger->critical($t->getMessage());
            return false;
        }

    }

    /**
     * @param $subject
     * @param $customerId
     */
    private function updateCouponUsages($subject, $customerId)
    {
        $this->coupon->load($subject->getCouponCode(), 'code');
        if ($this->coupon->getId()) {
            $this->couponUsage->updateCustomerCouponTimesUsed($customerId, $this->coupon->getId());
        }
    }

    /**
     * @param $order
     * @param $customerId
     * @throws Exception
     */
    private function addCustomerRule($order, $customerId)
    {
        $rules = explode(',', $order->getAppliedRuleIds());
        if (count($rules)) {
            foreach ($rules as $rule) {
                $this->updateCustomerRuleUsages($customerId, $rule);
            }
        }
    }

    /**
     * @param $customerId
     * @param $ruleId
     * @throws Exception
     */
    private function updateCustomerRuleUsages($customerId, $ruleId)
    {
        $ruleCustomer = $this->ruleCustomerFactory->create();
        $ruleCustomer->loadByCustomerRule($customerId, $ruleId);
        $ruleCustomer->setCustomerId($customerId)->setRuleId($ruleId)->setTimesUsed(1);
        $ruleCustomer->save();
    }

    /**
     * @param OrderInterface $order
     * @param AddressInterface $address
     * @return bool
     */
    private function needToSaveAddress(
        OrderInterface $order,
        AddressInterface $address
    ): bool {
        /** @var OrderAddressInterface|null $orderAddress */
        $orderAddress = null;
        if ($address->isDefaultBilling()) {
            $orderAddress = $order->getBillingAddress();
        } elseif ($address->isDefaultShipping()) {
            $orderAddress = $order->getShippingAddress();
        }
        if ($orderAddress) {
            $quoteAddressId = $orderAddress->getData('quote_address_id');
            if ($quoteAddressId) {
                /** @var QuoteAddress $quote */
                $quote = $this->addressFactory->create()
                    ->load($quoteAddressId);
                if ($quote && $quote->getId()) {
                    return (bool)(int)$quote->getData('save_in_address_book');
                }
            }

            return true;
        }

        return false;
    }


}

